unit BothSides.Import.Core;

interface

uses SysUtils, Classes, System.RTTI, Generics.Collections, Generics.Defaults, BothSides.Lists;

Type

  EImport = class( Exception )

  end;

  EImportSchema = class( EImport )

  end;

  FieldNameAttribute = class( TCustomAttribute )
  private
    FName: String;
  public
    constructor Create( AName: String );
    property Name : String read FName;
  end;

  TFieldNameList = class( TAttributeList<FieldNameAttribute> )

  end;

  TAttributeArray = TArray<TCustomAttribute>;
  TFieldArray = TArray<TRttiField>;

  CSVColumnAttribute = class( FieldNameAttribute )
  private
    FColumnIndex: Integer;
    FRequired: Boolean;
  public
    constructor Create( AName : String;ARequired : Boolean = false;ColIdx : Integer = 0);
    property ColumnIndex: Integer read FColumnIndex;
    property Required : Boolean read FRequired;
  end;

  TCSVSchemaItem = class
  private
    FColumnIndex: Integer;
    FColumnNames: TListOfStrings;
    procedure SetColumnIndex(const Value: Integer);
  public
    constructor Create;
    destructor Destroy; override;
    property ColumnNames: TListOfStrings read FColumnNames;
    property ColumnIndex : Integer read FColumnIndex write SetColumnIndex;
  end;

  TCSVSchemaItems = class( TObjectList<TCSVSchemaItem> )
  private
    FMaxColumnProcessed : Integer;
  protected
    function NewSchemaItem : TCSVSchemaItem;
  public
    procedure AddNewItem( AnAttr: CSVColumnAttribute );overload;
    procedure AddNewItem( AName: String;AnIdx : Integer );overload;
    procedure AddNewItem( AnAttr: CSVColumnAttribute;FieldNameList: TFieldNameList );overload;
  end;

  TCSVSchemaBuilder = class
  private
    FSchemaItems : TCSVSchemaItems;
    FNumRequired : Integer;
    FLineSeparator: String;
    procedure SetLineSeparator(const Value: String);
  protected
    procedure BuildSchemaFromClass( AClassTypeObject : TRttiType );
    procedure BuildSchemaFromRecord( ARecordType : TRttiRecordType );
    procedure BuildSchemaFromInterface( AnIntfType : TRttiInterfaceType );
    procedure BuildSchemaItemFromAttributes( AttrArray : TAttributeArray );
  public
    constructor Create;
    destructor Destroy; override;
    procedure BuldSchema<T>( SchemaTemplate: T );
    property NumRequired : Integer read FNumRequired;
    property SchemaItems : TCSVSchemaItems read FSchemaItems;
    property LineSeparator : String read FLineSeparator write SetLineSeparator;
  end;

  TCSVImporter = class;

  TCSVLine = class
  private
    FLineItems : TStringList;
    FImporter : TCSVImporter;
    function GetColumnsCount: Integer;
  protected
    property LineItems : TStringList read FLineItems;
  public
    constructor Create( Importer : TCSVImporter; ALine: String );
    destructor Destroy; override;
    property ColumnsCount : Integer read GetColumnsCount;
  end;

  TCSVHeader = class( TCSVLine )
  public
    // Check the right column name: this could be an exact match or it may have a
    // different case: to cater for the latter we allow an extra variable which will be
    // not empty. We then check that and, if we find it, then we re-assign the column
    // name to be correct.
    function ColumnFound( AName : String;out CorrectColumnName : String ) : Boolean;
  end;

  TCSVImporter = class
  private
    FSchemaBuilder : TCSVSchemaBuilder;
    // if file isn't loaded, we can't read next line
    FFileIsLoaded: Boolean;
    // We can auto-detect based off schema
    FColumnsOnFirstLine : Boolean;
    FCSVBuffer : TStringList;
    FNextLineIdx : Integer;
    FSeparator : String;
  protected
    procedure LoadLineInClass<T:class>( Buffer : T );
    procedure LoadLineInRecord<T:record>( Buffer: T);
  public
    constructor Create;
    procedure LoadCsv<T>( FileName : String;FieldSeparator : String;SchemaType : T );
    function NextLine<T>( var Buffer : T ) : Boolean;overload;
    function NextLine: Boolean;overload;
    procedure LoadLineValues<T>( var Buffer : T );
    procedure DoneImporting;
    property ActiveSchemaBuilder : TCSVSchemaBuilder read FSchemaBuilder;
  end;

{
  NOTES
  -----------------------------------------------------------------------------
  This import example is very opinionated and it uses exceptions in the way
  they should be used: to report things that we do not expect. Besides that,
  the golden rule remains that all public/published properties and methods
  should be allowed to be called indepndently of one another.
}

var RTTIContext : TRttiContext;
implementation

uses System.TypInfo;

{ FieldNameAttribute }

constructor FieldNameAttribute.Create(AName: String);
begin
  inherited Create;
  FName := AName;
end;

{ CSVColumn }

constructor CSVColumnAttribute.Create(AName: String;ARequired: Boolean; ColIdx: Integer);
begin
  inherited Create(AName);
  FColumnIndex := ColIdx;
  FRequired := ARequired;
end;

{ TCSVSchemaItem }

constructor TCSVSchemaItem.Create;
begin
  inherited Create;
  FColumnNames := TListOfStrings.Create;
  FColumnNames.Capacity := 10;
end;

destructor TCSVSchemaItem.Destroy;
begin
  FColumnNames.Free;
  inherited;
end;

procedure TCSVSchemaItem.SetColumnIndex(const Value: Integer);
begin
  FColumnIndex := Value;
end;

{ TCSVSchemaItems }

procedure TCSVSchemaItems.AddNewItem(AnAttr: CSVColumnAttribute);
begin
  AddNewItem( AnAttr.Name,AnAttr.ColumnIndex );
end;

procedure TCSVSchemaItems.AddNewItem(AName: String; AnIdx: Integer);
var NewItem : TCSVSchemaItem;
begin
  NewItem := NewSchemaItem;
  NewItem.ColumnNames.Add( AName );
  NewItem.ColumnIndex := AnIdx;
  // Only trigger if there is at least an element.
  if ( NewItem.ColumnIndex = 0 ) And ( Count > 0 ) then
    begin
      NewItem.ColumnIndex := FMaxColumnProcessed +1;
      Inc( FMaxColumnProcessed );
    end;
  Add( NewItem );
end;

procedure TCSVSchemaItems.AddNewItem(AnAttr: CSVColumnAttribute;
  FieldNameList: TFieldNameList);
var FldAttr : FieldNameAttribute;
    LastItemIdx : Integer;
begin
  AddNewItem( AnAttr );
  // What happens here is that we have just added a new column. Therefore, the
  // one we want is at Count -1
  LastItemIdx := Count -1;
  for FldAttr in FieldNameList do
    begin
      Items[ LastItemIdx ].ColumnNames.Add( FldAttr.Name );
    end;
end;

function TCSVSchemaItems.NewSchemaItem: TCSVSchemaItem;
begin
  Result := TCSVSchemaItem.Create;
end;

{ TCSVSchemaBuilder }

procedure TCSVSchemaBuilder.BuildSchemaFromClass(
  AClassTypeObject: TRttiType);
var PropList : TList<TRttiProperty>;
    AProperty : TRttiProperty;
    AClassProperties : TArray<TRttiProperty>;
    AttrArray : TAttributeArray;
begin
  // We are building a schema from a class type.
  // First thing, we take all the public and published properties and put
  // them into a list.
  AClassProperties := AClassTypeObject.GetProperties;
  PropList := TList<TRttiProperty>.Create;
  // Start with 20 because that should cover most CSV files.
  PropList.Capacity := 20;
  for AProperty in AClassProperties do
    begin
      // Only public/published property that aren't read-only can be used
      if ( AProperty.Visibility in [ mvPublic,mvPublished ] ) And AProperty.IsWritable then
         PropList.Add( AProperty );
    end;
  if PropList.Count = 0 then
    begin
      raise EImportSchema.Create('No public or published properties to create a valid schema for class '+AClassTypeObject.Name);
    end;
  // OK so we have a list of public/published properties to work with.
  // All we need to do now is to examine for column and attributes and see if
  // there is anything that we can use. Our assumption here is that if there is
  // more than one attribute, then one is of type CSVColumnAttribute and the
  // others of type FieldNameAttribute.
  for AProperty in PropList do
    begin
      AttrArray := AProperty.GetAttributes;
      // The attributes work always the same way, what changes is how and where we find them!
      BuildSchemaItemFromAttributes( AttrArray );
    end;
  PropList.Free;
end;

procedure TCSVSchemaBuilder.BuildSchemaFromInterface(
  AnIntfType: TRttiInterfaceType);
var Properties: TArray<TRttiProperty>;
    AProperty : TRttiProperty;
    AttrArray : TAttributeArray;
begin
  // Unlike classes, all interface properties are public!
  // Because of this, we don't need a special list and we can directly roll
  // with the punches
  Properties := AnIntfType.GetProperties;
  if Length( Properties ) = 0 then
    raise EImportSchema.Create('No properties defined for interface '+AnIntfType.Name);
  for AProperty in Properties do
    begin
      AttrArray := AProperty.GetAttributes;
      BuildSchemaItemFromAttributes( AttrArray );
    end;
end;

procedure TCSVSchemaBuilder.BuildSchemaFromRecord(ARecordType: TRttiRecordType);
var AttrArray : TAttributeArray;
    Fields : TFieldArray;
    AField : TRttiField;
begin
  Fields := ARecordType.GetDeclaredFields;
  if Length( Fields ) = 0 then
    raise EImportSchema.Create('No fields have been defined for record'+ARecordType.Name);
  for AField in Fields do
    begin
      AttrArray := AField.GetAttributes;
      // The attributes work always the same way, what changes is how and where we find them!
      BuildSchemaItemFromAttributes( AttrArray );
    end;
end;

procedure TCSVSchemaBuilder.BuildSchemaItemFromAttributes(
  AttrArray: TAttributeArray);
var Attr : TCustomAttribute;
    FieldNames : TFieldNameList;
    FNAttr : FieldNameAttribute;
    CSVAttr: CSVColumnAttribute;
begin
  // No attributes means it should be skipped.
  if Length(AttrArray) > 0 then
    begin
       if Not ( AttrArray[ 0 ] is CSVColumnAttribute ) then
         raise EImportSchema.Create('First attribute MUST be CSV column');
       CSVAttr := CSVColumnAttribute(AttrArray[ 0 ]);
       if CSVAttr.Required then
         Inc( FNumRequired );
       FieldNames := TFieldNameList.Create;
       try
         // More than 5 aliases? Come on now...
         FieldNames.Capacity := 6;
         for Attr in AttrArray do
           begin
             if Attr is FieldNameAttribute then
               begin
                 FNAttr := FieldNameAttribute( Attr );
                 FieldNames.Add( FNAttr );
               end;
           end;
           FSchemaItems.AddNewItem( CSVAttr,FieldNames );
       finally
         FieldNames.Free;
       end;
    end;
end;

procedure TCSVSchemaBuilder.BuldSchema<T>(SchemaTemplate: T);
var AType : TRttiType;
begin
  FSchemaItems.Clear;
  FNumRequired := 0;
  // Get type
  AType := RTTIContext.GetType(  TypeInfo(T) );
  // if it's one of the correct types
  if AType.TypeKind in [ tkRecord,tkClass,tkInterface ] then
    begin
      if AType.TypeKind = tkClass then
        BuildSchemaFromClass( AType )
      else if AType.TypeKind = tkRecord then
             begin
               if AType.IsRecord then
                 BuildSchemaFromRecord( AType.AsRecord );
             end
      else
        begin
          // If it's not a record or class, but it's still good, then it must be an interface type!
          BuildSchemaFromInterface( AType as TRttiInterfaceType );
        end;
    end
  else
    begin
      raise EImportSchema.Create('Cannot create schema by type '+AType.Name );
    end;
  if FSchemaItems.Count = 0 then
    raise EImportSchema.Create('Schema is missing critical information. Cannot proceed for '+AType.Name);
end;

constructor TCSVSchemaBuilder.Create;
begin
  inherited Create;
  FSchemaItems := TCSVSchemaItems.Create( True );
end;

destructor TCSVSchemaBuilder.Destroy;
begin
  FSchemaItems.Free;
  inherited;
end;

procedure TCSVSchemaBuilder.SetLineSeparator(const Value: String);
begin
  FLineSeparator := Value;
end;

{ TCSVImporter }

constructor TCSVImporter.Create;
begin
  inherited Create;
  FSchemaBuilder := TCSVSchemaBuilder.Create;
  FCSVBuffer := TStringList.Create;
end;

procedure TCSVImporter.DoneImporting;
begin
  FCSVBuffer.Clear;
  FFileIsLoaded := False;
end;

procedure TCSVImporter.LoadCsv<T>(FileName, FieldSeparator: String;
  SchemaType: T);
var Item : TCSVSchemaItem;
    ColName, RealColName : String;
    CSVHeader : TCSVHeader;
    CSVLine : TCSVLine;
    NumOfColumnsFound : Integer;
begin
  try
    FCSVBuffer.LoadFromFile( FileName );
    // Becomes true only if it's actually loaded
    FFileIsLoaded := True;
    FSchemaBuilder.BuldSchema<T>( SchemaType );
    FSchemaBuilder.LineSeparator := FieldSeparator;
    FNextLineIdx := 0;
    NumOfColumnsFound := 0;
    // Now that the schema is built, we need to check if the first line is made of columns.
    for Item in FSchemaBuilder.SchemaItems do
      begin
        CSVHeader := TCSVHeader.Create( Self,FCSVBuffer.Strings[ 0 ] );
        for ColName in Item.ColumnNames do
          begin
            if CSVHeader.ColumnFound( ColName,RealColName ) then
              begin
                if Not RealColName.IsEmpty then
                  begin
                    // Just assign it as first element because it's easier
                    Item.ColumnNames.Items[ 0 ] := RealColName;
                  end;
                Inc( NumOfColumnsFound );
                // Found it, get out.
                break;
              end;
          end;
      end;
    FColumnsOnFirstLine := NumOfColumnsFound > 0;
    if FSchemaBuilder.NumRequired > NumOfColumnsFound then
      raise EImportSchema.Create('Could not find as many required columns in '+FileName+' as the schema suggests there should be.');
    if FColumnsOnFirstLine then
      Inc( FNextLineIdx );

  except
    // we don't care about what went wrong, whatever it is we simply can't
    // consider the file loaded
    FFileIsLoaded := False;
    // re-raise the exception so it doesn't get lost.
    raise;
  end;
end;

procedure TCSVImporter.LoadLineInClass<T>(Buffer: T);
begin

end;

procedure TCSVImporter.LoadLineInRecord<T>(Buffer: T);
begin

end;

procedure TCSVImporter.LoadLineValues<T>(var Buffer: T);
var AType: TRttiType;
    ObjectBuffer : TObject;
begin
  AType := RTTIContext.GetType( TypeInfo(T) );
  if AType.TypeKind = tkClass then
    begin
       ObjectBuffer := TObject(Buffer);
       LoadLineInClass<TObject>( ObjectBuffer );
    end
  else
    if AType.TypeKind = tkRecord then
      LoadLineInRecord<T>( Buffer );
end;

function TCSVImporter.NextLine: Boolean;
begin
  if FFileIsLoaded then
    begin
      Inc( FNextLineIdx );
      Result := FNextLineIdx < FCSVBuffer.Count;
    end
  else Result := False;
end;

function TCSVImporter.NextLine<T>(var Buffer: T): Boolean;
begin
  if Not FFileIsLoaded then
  begin
    Result := False;
    Exit;
  end;
  LoadLineValues<T>(Buffer);
  Result := NextLine;
end;

{ TCSVLine }

constructor TCSVLine.Create(Importer: TCSVImporter; ALine: String);
begin
  inherited Create;
  FImporter := Importer;
  FLineItems := TStringList.Create;
  FLineItems.LineBreak := FImporter.ActiveSchemaBuilder.LineSeparator;
  FLineItems.Text := ALine;
end;

destructor TCSVLine.Destroy;
begin
  FLineItems.Free;
  inherited;
end;

function TCSVLine.GetColumnsCount: Integer;
begin
  Result := FLineItems.Count;
end;

{ TCSVHeader }

function TCSVHeader.ColumnFound(AName: String;out CorrectColumnName : String): Boolean;
var AColumnName : String;
begin
  Result := LineItems.IndexOf( AName ) >= 0;
  CorrectColumnName := '';
  if Not Result then
    begin
      for AColumnName in LineItems do
        begin
          Result := SameText( AName,AColumnName );
          if Result then
            CorrectColumnName := AColumnName;
        end;
    end;
end;

initialization
begin
  RTTIContext := TRttiContext.Create;
end;

finalization
begin
  RTTIContext.Free;
end;

end.
